/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <limits.h>

/**
The devssim functions as the root coordinator for a simulation.
In generally, it should not be sub-classed, except to override the
halt () method.
*/

class devssim: public object {

public:		
/// Creates and initializes a simulator for the model with initial time t0.
		devssim (devs* model, adevs::stime_t t0 = 0.0);
/**
   Step the simulator by dt units of time.  Throws a ScheduleException
   if dt < 0 or if (timeCurrent() + dt) > timeNext ().
*/
		void step (stime_t dt);
/**
   Steps the simulator after applying the port_values in the 
   input list to the top level model.  The input list is 
   emptied by the step method.  Input objects delivered to the
   simulator are adopted by the simulator (i.e. the caller
   should not tamper with the input after the step 
   call has been made).
*/
		void step (stime_t dt, indexed& input);
/**
   Computes the output of the top level model at tN.  The indexed
   list provided to the method is filled with port_value objects,
   one for each item on the output ports.  The port_value objects
	and their value objects are owned by the caller.
   <p> The state of the model is not changed by called this method, 
   provided that the output_func method does not cause a state change.
	Garbage collection is not performed at the end of the output computation,
	and so objects outputted by the output_func() are not destroyed.  These
	objects ARE deleted when the next state change is computed (i.e. via
	step or run).
   <p>This method fails with a CloneException if any of the output
   objects do not support the clone() method.
*/
		void computeNextOutput (indexed& result);
/// Run the simulator.
		inline void run () { run (INFINITY, UINT_MAX); }
/// Run the simulator until tN > t_max
		inline void run (stime_t t_max) { run (t_max, UINT_MAX); }
/// Run the simulator through i cycles
		inline void run (unsigned i) { run (INFINITY, i); }
/// Run the simulator through to t_max or max_cycles (whichever comes first).
		void run (stime_t t_max, unsigned max_cycles);
/// Returns the current simulation time.
		stime_t timeCurrent ();
/// Returns the time of next event.
		stime_t timeNext ();
/// Returns the time of of the last event.
		stime_t timeLast ();
/// Returns true if tN = INFINITY, false otherwise.
		bool isPassive ();
/**
Resets the simulator.  If model = true, reset the model attached
to the simulator as well.
*/
		virtual void reset (stime_t t0, bool reset_model = true);
		virtual void reset (bool reset_model = true);
/**
Returns true if the simulator should stop, false otherwise.  This method
is called by the root coordinator/simulator at the end of each simulation
cycle.
*/
		virtual bool halt ();
/**
Deletes the root simulator and all of its components.
*/
		virtual ~devssim ();
/**
Set the thread limit for running parallel simulations.  A value of 0
will result in a sequentail run. (default is 0)
*/
		static inline void setThreadLimit (int limit) {
			psimulator::setThreadLimit (limit);
			}

protected:	
		devs* model;

private:	
		stime_t t;
		simulator* sim;
		indexed garbage;

		void compute_cycle (stime_t t);

};
