/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs/adevs.h>
using namespace adevs;

void adevs::do_conservative_execution (pDEVSsimulator* sim) {

// Simulate until we've used up the available lookahead
// Model should not produce output during this time,
// so the output function is not called.
while (sim->model->timeNext () < sim->t_thread) 
   sim->_computeTransition (sim->model->timeNext());
   
}

void adevs::sim_compute_next_output (void* arg) {

pDEVSsimulator* sim = (pDEVSsimulator*)arg;
// Advance state of model until we've reached a 
// (potential) output producing state
do_conservative_execution (sim);
// Compute the output for that state
sim->_computeNextOutput (sim->t_thread);
// Unlock the mutex (locked by sim before function call)
sim->mtx.unlock ();

}

void adevs::sim_compute_io (void* arg) {

pDEVSsimulator* sim = (pDEVSsimulator*)arg;
// Advance state of model until we've reached a 
// (potential) output producing state
do_conservative_execution (sim);
// Compute the output for that state
sim->_computeIO (sim->t_thread);
// Unlock the mutex (locked by sim before function call)
sim->mtx.unlock ();

}

void adevs::sim_compute_transition (void* arg) {

pDEVSsimulator* sim = (pDEVSsimulator*)arg;
// Advance state of model until we've used up 
// the available lookahead
do_conservative_execution (sim);
// Compute the system state for t_thread
sim->_computeTransition (sim->t_thread);
// Unlock the mutex (locked by sim before function call)
sim->mtx.unlock ();

}

pDEVSsimulator::pDEVSsimulator ():
DEVSsimulator (simulator::PSIM) { 

t_thread = 0.0;
thr = NULL;

}

void pDEVSsimulator::computeNextOutput (stime_t t) {

t_thread = t;
if (thr == NULL)
	thr = getThread();
if (thr == NULL) {
	do_conservative_execution (this);
	DEVSsimulator::computeNextOutput(t);
	}
else {
	mtx.lock ();
	thr->run (sim_compute_next_output, this);
	}

}

void pDEVSsimulator::_computeNextOutput (stime_t t) {

DEVSsimulator::computeNextOutput (t);

}

void pDEVSsimulator::computeIO (stime_t t) {

t_thread = t;
if (thr == NULL)
	thr = getThread();
if (thr == NULL) {
	do_conservative_execution (this);
	DEVSsimulator::computeIO(t);
	}
else {
	mtx.lock ();
	thr->run (sim_compute_io, this);
	}

}

void pDEVSsimulator::_computeIO (stime_t t) {

DEVSsimulator::computeIO (t);

}

void pDEVSsimulator::computeTransition (stime_t t) {

t_thread = t;
if (thr == NULL)
	thr = getThread();
if (thr == NULL) {
	do_conservative_execution (this);
	DEVSsimulator::computeTransition(t);
	}
else {
	mtx.lock ();
	thr->run (sim_compute_transition, this);
	}

}

void pDEVSsimulator::_computeTransition (stime_t t) {

DEVSsimulator::computeTransition (t);

}

stime_t pDEVSsimulator::timeNext () {

stime_t result = model->timeNext() + model->lookahead ();
if (model->timeNext() == INFINITY)
   result = INFINITY;
else if (result < model->timeNext())
   result = model->timeNext (); 
return result;

}

stime_t pDEVSsimulator::sigma () {

stime_t result = (model->lookahead() + model->timeNext()) - model->timeCurrent();
if (model->sigma() == INFINITY)
   result = INFINITY;
return result;

}

void pDEVSsimulator::wait () {

if (thr != NULL) {
	mtx.lock ();
	mtx.unlock ();
	returnThread(thr);
	thr = NULL;
	}

}

pDEVSsimulator::~pDEVSsimulator () { }
