/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs/adevs.h>
using namespace adevs;

/**
Structures for tracking member index information.
**/
// Node in a linked list of index numbers
class _index_info {
	public:
		_index_info* next;
		unsigned value;
		inline _index_info() {}
};

// Linked list of index numbers
class _member_info: public object {
	private:	
		unsigned instance_count;
		// index information is stored as a linked list
		_index_info* list;
	public:		
		inline _member_info():object() { 
			instance_count = 0;
			list = NULL;
		}
		_member_info(const _member_info &src);
		void addInstance(unsigned index);
		void removeInstance(unsigned index);
		inline unsigned instanceCount() { return instance_count; }
		// do not call if instanceCount is 0
		inline unsigned getIndex() { return list->value; }
		~_member_info();
};

_member_info::_member_info(const _member_info &src):
object(src) {
	instance_count = src.instance_count;
	if (src.list == NULL) list = NULL;
	else {
		list = new _index_info();
		_index_info *t = list;
		_index_info *s = src.list;
		t->value = s->value;
		while (s->next != NULL) {
			t->next = new _index_info();
			t = t->next;
			s = s->next;
			t->value = s->value;
			}
		t->next = NULL;
		}
} // end of copy constructor
 
void _member_info::addInstance(unsigned index) {
	_index_info* head;
	head = new _index_info();
	head->value = index;
	head->next = list;
	list = head;
	instance_count++;
}

// should only be called for valid indices
void _member_info::removeInstance(unsigned index) {
	_index_info* pos = list;
	_index_info* last;
	while (pos->value != index) {
		last = pos;
		pos = pos->next;
		}
	if (pos == list) list = pos->next;
	else last->next = pos->next; 
	delete pos;
	instance_count--;
}

_member_info::~_member_info() {
	_index_info* last;
	_index_info* pos = list;
	while (pos != NULL) {
		last = pos;
		pos = pos->next;
		delete last;
		}
}

// end of membership info implementation

// begin collection implementation

collection::collection():
indexed() {
	_table = new hashtable(DEFAULT_CONTAINER_CAPACITY * 2);
}

collection::collection(unsigned capacity, unsigned grow):
indexed(capacity, grow) {
	_table = new hashtable(capacity * 2);
}

collection::collection(const collection &src):
indexed(src.getCapacity()) {
	_table = new hashtable (src.getCapacity()*2);
	for (unsigned i = 0; i < src.getSize(); i++)
		add (src.getObj(i));
} // end of copy constructor

void collection::swap(unsigned i, unsigned j) {
	_member_info* item_i = (_member_info*)_table->find(*this->getObj(i));
	_member_info* item_j = (_member_info*)_table->find(*this->getObj(j));
	item_i->removeInstance(i);
	item_i->addInstance(j);
	item_j->removeInstance(j);
	item_j->addInstance(i);
	indexed::swap(i,j);
} // end of swap

void collection::empty() {
	for (unsigned i = 0; i < getSize(); i++) 
		_table->erase (*(getObj(i)));
	indexed::empty();
} // end of empty

void collection::add(object *obj) {
	indexed::add (obj);
	_member_info* item =(_member_info*)_table->find(*obj);
	if (item == NULL) {
		item = new _member_info();
		_table->insert(obj,item);
		}
	item->addInstance(getSize() - 1);
} // end of add

object* collection::remove(unsigned index) {
	object* to_remove = this->getObj(index);
	_member_info* info;
	unsigned last_index = getSize() - 1;
	object* last = this->getObj(last_index);
	info = (_member_info*)_table->find(*last);
	info->removeInstance(last_index);
	info->addInstance(index);
	info = (_member_info*)_table->find(*to_remove);
	info->removeInstance(index);
	if (info->instanceCount () == 0) {
		_table->remove(*to_remove);
		delete info;
		}
	return indexed::remove(index);
} // end of remove

void collection::removeAll() {
	for (unsigned i = 0; i < getSize(); i++) 
		_table->erase(*(getObj(i)));
	indexed::removeAll();
}

unsigned collection::instanceCount(const object &obj) const {
	_member_info* item = (_member_info*)_table->find(obj);
	if (item == NULL) return 0;
	return item->instanceCount();
} // end of instanceCount
 
unsigned collection::indexOf(const object &obj) const {
	_member_info* item = (_member_info*)_table->find (obj);
	return item->getIndex ();
} // end of indexOf

collection::~collection() {
	for (unsigned i = 0; i < getSize(); i++)
		_table->erase (*(getObj(i))); 
	delete _table;
} // end of class collection
