/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <adevs/adevs.h>
using namespace adevs;

hashtable::hashtable():
object() {
	keys = 0;
	this->buckets = DEFAULT_CONTAINER_CAPACITY;
	this->loadlimit = DEFAULT_HASH_LOAD;
	table = new indexed* [buckets];
	for (unsigned int i = 0; i < this->buckets; i++) 
		table[i] = NULL;
} // end of constructor

hashtable::hashtable(unsigned int buckets, float loadlimit):
object() {
	this->buckets = buckets;
	this->loadlimit = loadlimit;
	keys = 0;
	table = new indexed* [buckets];
	for (unsigned int i = 0; i < this->buckets; i++) 
		table[i] = NULL;
} // end of constructor

hashtable::hashtable(const hashtable &src) {
	buckets = src.buckets;
	loadlimit = src.loadlimit;
	keys = src.keys;
	table = new indexed* [buckets];
	for (unsigned int i = 0; i < this->buckets; i++) {
		if (src.table[i] != NULL) {
			table[i] = new indexed (src.table[i]->getSize(), HASH_LIST_CAPACITY);
			for (unsigned j = 0; j < src.table[i]->getSize (); j++) {
				table[i]->add(new _bucket(*(_bucket*)src.table[i]->getObj(j)));
				}
			}
		else
			table[i] = NULL;
		}
} // end of copy constructor

object* hashtable::insert(object *key, object *value) {
	unsigned int code = key->hashcode() % buckets;
	indexed* l = (indexed*)table[code];
	object* last = NULL;
	// if other entries, look for previous existance of key
	if (l != NULL) { 
		for (unsigned i = 0; i < l->getSize(); i++) {
			_bucket* b = (_bucket*)l->getObj(i);
			if (key->operator==(*b->key)) {
				last = b->value;
				b->value = value;
				return last;
				}
			}
		}
	// if no other entries at this index
	else if (l == NULL) 
		l = table[code] = new indexed(HASH_LIST_CAPACITY, HASH_LIST_GROWTH);
	// if no previous instance, add a new bucket
	l->add(makeBucket(key, value));
	keys++;
	if ((float)buckets * loadlimit <= (float)keys) 
		rehash();
	return NULL;
} // end of insert

object* hashtable::erase(const object& key) {
	unsigned int code = key.hashcode() % buckets;
	indexed* l = table[code];
	object* last_key = NULL;
	if (l == NULL) return NULL;
	for (unsigned i = 0; i < l->getSize(); i++) {
		_bucket *b = (_bucket*)l->getObj(i);
		if (key == *(b->key)) {
			last_key = b->key;
			l->remove(i);
			delete b->value;
			freeBucket(b);
			keys--;
			break; 
			}
		}
	return last_key;
} // end of erase

object* hashtable::remove(const object &key) {
	unsigned int code = key.hashcode() % buckets;
	indexed* l = table[code];
	object* last_key = NULL;
	if (l == NULL) return NULL;
	for (unsigned i = 0; i < l->getSize(); i++) {
		_bucket *b = (_bucket*)l->getObj(i);
		if (key == *(b->key)) {
			last_key = b->key;
			l->remove(i);
			freeBucket(b);
			keys--;
			break; 
			}
		}
	return last_key;
} // end of remove

object* hashtable::find(const object &key) const {
	unsigned int code = key.hashcode() % buckets;
	indexed *l = (indexed*)table[code];
	if (l == NULL) return NULL;
	for (unsigned i = 0; i < l->getSize(); i++) {
		_bucket *b = (_bucket*)l->getObj(i);
		if (key == *(b->key)) return b->value;
		}
	return NULL;
} // end of find

object** hashtable::getKeys() const {
	object** keytable = new object* [keys];
	unsigned int keycounter = 0;
	for (unsigned int i = 0; keycounter < keys; i++) {
		indexed* l = table[i];
		if (l != NULL) {
			for (unsigned j = 0; j < l->getSize(); j++) {
				_bucket* b = (_bucket*)l->getObj(j);
				keytable[keycounter] = b->key;
				keycounter++;
				}
			}
		}
	return keytable;
} // end of getKeys

void hashtable::clear() {
	for (unsigned int i = 0; i < buckets; i++) {
		indexed* l = table[i]; 
		if (l != NULL) l->empty();
		}
	keys = 0;
} // end of clear

void hashtable::rehash() {
	unsigned int rbuckets = buckets * 3;
	indexed** rtable = new indexed* [rbuckets]; 
	for (unsigned int i = 0; i < rbuckets; i++) rtable[i] = NULL;
	unsigned int keycounter = 0;
	for (unsigned int i = 0; keycounter < keys; i++) {
		indexed* l = table[i];
		if (l != NULL) {
		for (unsigned j = 0; j < l->getSize(); j++) {
			_bucket* b = (_bucket*)l->getObj(j);
			unsigned code = b->key->hashcode() % rbuckets;
			indexed* lr = rtable[code];
			if (lr == NULL) 
				lr = rtable[code] = new indexed(l->getSize(),HASH_LIST_GROWTH);
			lr->add(b);
			}
		keycounter += l->getSize();
		delete l;
		}
	}
	buckets = rbuckets;
	delete [] table;
	table = rtable;
} // end of rehash

hashtable::~hashtable() {
	for (unsigned int i = 0; i < buckets; i++) {
		indexed* l = table[i];
		if (l != NULL) {
			l->empty();
			delete l;
			}
		}
	bucket_pool.empty();
	delete [] table;
}

hashtable::_bucket* hashtable::makeBucket (object* key, object* value) {
	if (bucket_pool.getSize() == 0)
		return new _bucket(key,value);
	else {
		_bucket* b = (_bucket*)bucket_pool.remove(0);
		b->key = key;
		b->value = value;
		return b;
		}
}

// end of class hashtable
