/***************
adevs, a library supporting discrete event simulation
Copyright (C) 1999, 2000, 2001 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include <float.h>
#include <adevs/adevs.h>
using namespace adevs;

const priority_t pqueue::min_priority = (double)(-DBL_MIN);

// an object wrapper around a unsigned int
class _uint: public object {
	public:		
		unsigned value;
		inline _uint(unsigned value):object() { this->value = value; }
};

// start of pqueue implementation
pqueue::pqueue ():
object() {
	this->capacity = DEFAULT_CONTAINER_CAPACITY;
	this->grow = DEFAULT_CONTAINER_GROWTH;
	heap = new _element* [capacity];
	heap[0] = new _element(NULL, min_priority);
	members = new hashtable(capacity);
}

pqueue::pqueue(unsigned capacity, unsigned grow):
object() {
	this->capacity = capacity;
	this->grow = grow;
	heap = new _element* [capacity];
	heap[0] = new _element(NULL, min_priority);
	members = new hashtable(capacity);
}

void pqueue::insert(object *obj, priority_t priority) {
	if (members->keyCount()+1 == capacity) 
		enlarge ();
	unsigned i = percolate_up(members->keyCount()+1,priority);
	heap[i] = new _element(obj,priority);
	members->insert(obj,new _uint (i));
}

object* pqueue::removeMinimum() {
	unsigned i;
	_element *min, *last;
	if (members->keyCount() == 0) return NULL;
	min = heap[1];
	last = heap[members->keyCount()];
	delete members->find(*min->item);
	members->remove(*min->item);
	object* result = min->item;
	delete min;
	if (members->keyCount() > 0) {
		i = percolate_down(1,last->priority);
		heap[i] = last;
		_uint *last_index = (_uint*)members->find(*last->item);
		last_index->value = i;
		}
	return result;
}

void pqueue::adjust(const object &item, priority_t priority) {
	_uint *tomove = (_uint*)members->find(item);
	_element *e = heap[tomove->value];
	if (e->priority > priority)
		tomove->value = percolate_up(tomove->value, priority);
	else if (e->priority < priority)
		tomove->value = percolate_down(tomove->value, priority);
	else
		return;
	heap[tomove->value] = e;
	e->priority = priority;
}

priority_t pqueue::minPriority() const {
	if (members->keyCount() == 0) return min_priority;
	else return heap[1]->priority;
}

priority_t pqueue::getPriority(const object &obj) const {
	_uint *i = (_uint*)members->find(obj);
	return heap[i->value]->priority;
}

bool pqueue::isEnqueued(const object &obj) const {
	return (members->find(obj) != NULL);
}

unsigned pqueue::percolate_up(unsigned index, priority_t priority) {
	_uint* member_index;
	unsigned i = index;
	while (heap[i/2]->priority > priority) {
		heap[i] = heap[i/2];
		member_index = (_uint*)members->find (*heap[i]->item);
		member_index->value = i;
		i /= 2;
		}
	return i;
}

unsigned pqueue::percolate_down(unsigned index, priority_t priority) {
	_uint* member_index;
	unsigned i, child;
	unsigned size = members->keyCount ();
	for (i = index; i*2 <= size; i = child) {
		child = i*2;
		if (child != size && heap[child+1]->priority < heap[child]->priority) 
			child++;
		if (priority > heap[child]->priority) {
			heap[i] = heap[child];
			member_index = (_uint*)members->find(*heap[i]->item);
			member_index->value = i;
			}
		else break;
		}
	return i;
}

void pqueue::enlarge() {
	_element **rheap;
	rheap = new _element* [capacity + grow];
	for (unsigned i = 0; i < capacity; i++)
		rheap[i] = heap[i];
	capacity += grow;
	delete [] heap;
	heap = rheap;
}

pqueue::~pqueue() {
	object** keys = members->getKeys ();
	unsigned size = members->keyCount ();
	for (unsigned i = 0; i < size; i++) {
		_uint* index = (_uint*)members->find(*keys[i]);
		members->remove(*keys[i]);
		delete heap[index->value];
		delete index;
		}
	delete [] keys;
	delete members;
	delete heap[0];
	delete [] heap;
}
