/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "xblock.h"
#include <iostream.h>
using namespace adevs;

// Ball will only report changes in position in excess of the MOTION_RES value.
const double ball::MOTION_RES = 0.4;

// input and output ports
const port_t ball::collision = 0;
const port_t ball::motion = 1;
const port_t ball::destroyed = 2;
const port_t ball::stop = 3;

ball::ball ():atomic ("ball") {

status = OK;

addInputPort (stop);
addInputPort (collision);
addOutputPort (motion);
addOutputPort (destroyed);

}

ball::ball (const vector& v_init):atomic ("ball") {

status = OK;
this->v_init = v = v_init;

addInputPort (stop);
addInputPort (collision);
addOutputPort (motion);
addOutputPort (destroyed);

}

ball::ball (const circle& geom_init):atomic ("ball") {

status = OK;
this->geom_init = geom = geom_init;

addInputPort (stop);
addInputPort (collision);
addOutputPort (motion);
addOutputPort (destroyed);

}

ball::ball (const circle& geom_init, const vector& v_init):atomic ("ball") {

status = OK;
this->v_init = v = v_init;
this->geom_init = geom = geom_init;

addInputPort (stop);
addInputPort (collision);
addOutputPort (motion);
addOutputPort (destroyed);

}

void ball::setWorld (const rectangle& world_size) {

this->world_size = world_size;

}

void ball::setWorld (double x, double y, double width, double height) {

world_size.setPosition (x, y);
world_size.setSize (width, height);

}

void ball::reset () {

status = OK;
v = v_init;
geom_op = geom = geom_init;
dp.x = dp.y = 0;
hold (0);

}

void ball::delta_int () {

// if ball is destroyed, reset and start bouncing again.
if (status == DESTROYED) {
   geom_op = geom = geom_init;
   v = v_init;
   status = OK;
   }

// current position gets the last anticipated position
geom = geom_op;
// compute the time needed to move a significant distance at
// the current velocity.
stime_t dt = timeToDist (MOTION_RES, v);
// get the position change that will occur in that time
dp = distInTime (dt, v);
// make sure we haven't gone off the edge of the world
checkStatus ();
// compute the next position
geom_op.deltaPos (dp);
// hold for the time needed to move to the next position
hold (dt);

}

void ball::delta_ext () {

// update the position of the ball to reflect the elapsed time
geom.deltaPos (distInTime (elapsed (), v));

// passivate the ball (freeze it) if input is received on the
// stop port.
bag* values = inputOnPort (stop);

if (values->getSize () > 0) {
   passivate ();
   return;
   }

// check for a collision event directed at this ball.
values = inputOnPort (collision);
for (unsigned i = 0; i < values->getSize (); i++) {
   collisionEvent* cp = (collisionEvent*)values->getObj (i);
   // if a collision is detected, compute the balls resulting velocity
   if (cp->getTarget () == this) {
      const shape* s = cp->getGeometry ();
      v = elasticCollision (v, s->surfaceAngle (geom.getPosition ()));
      }
   }

// compute time to move significant distance at (new?) velocity
stime_t dt = timeToDist (MOTION_RES, v);
// compute change in position to occur in that time
dp = distInTime (dt, v);
// compute the next position
geom_op = geom;
geom_op.deltaPos (dp);
// hold until position is reached 
hold (dt);

}

void ball::delta_conf () { 

delta_int ();
delta_ext (); 

}

void ball::output_func () {

// if destroyed, output the balls new status
if (status != OK)
   output (destroyed, new statusEvent (this, status, &geom));

// output the balls updated position
output (motion, new motionEvent (this, v, &geom_op));

}

// destroy the ball if it goes outside of the bounding box
void ball::checkStatus () {

if (!world_size.overlaps (geom)) status = DESTROYED;

}

ball::~ball () { }
