/***********
Copyright (C) 1999, 2000  James Nutaro

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "message.h"
#include "process.h"
#include "channel.h"
#include "prog.h"
#include <adevs/adevs.h>
#include <iostream>
#include <algorithm>
#include <time.h>
#include <unistd.h>
using namespace adevs;

static rv r;

/**

Model of a processor.  Contains a process (distributed execution coordinator) and
the attached program.

*/
class processor: public staticDigraph {

public:
	processor (unsigned pid, unsigned num_procs, 
		stime_t max_proc_period, unsigned max_messages);

	inline unsigned numExecuted () const { return _prog->numExecuted (); }
	inline const message& execList (unsigned i) const { 
		return _prog->execList (i); }
	~processor ();

	static const port_t in;
	static const port_t out;

private:
	prog* _prog;

};

const port_t processor::in = 0;
const port_t processor::out = 1;

processor::processor (unsigned pid, unsigned num_procs, stime_t proc_period,
			unsigned max_messages):
staticDigraph ("processor") {

addInputPort (in);
addOutputPort (out);

process* p = new process (pid, num_procs);
unsigned msgs = (unsigned)r.uniform (1, (double)max_messages); 
_prog = new prog (pid, proc_period, msgs);
add (p); add (_prog);
couple (p, process::prog_out, _prog, prog::in);
couple (_prog, prog::out, p, process::prog_in);
couple (this, in, p, process::recv);
couple (p, process::send, this, out);

}

processor::~processor () { }

/**

Model of the distributed system.  Consists of several processors interconnected
by FIFO channels.

*/
class sys: public staticDigraph {

public:
	sys (unsigned num_procs, stime_t max_delay, 
	stime_t max_proc_period, unsigned max_messages);

	void report ();
	~sys ();

private:
	indexed lps;

};

sys::sys (unsigned num_procs, stime_t max_delay, 
	stime_t max_proc_period, unsigned max_messages)
:staticDigraph ("sys") {

for (unsigned i = 0; i < num_procs; i++) {
   processor* p = new processor (i, num_procs, max_proc_period, max_messages);
   lps.add (p);
   add (p);
   }

for (unsigned i = 0; i < num_procs; i++) {
   for (unsigned k = 0; k < num_procs; k++) {
      if (i != k) {
         processor* src = (processor*)lps[i];
         processor* snk = (processor*)lps[k];
         channel* c = new channel (i * k, max_delay);
         add (c);
         couple (src, processor::out, c, channel::in);
         couple (c, channel::out, snk, processor::in);
         }
      }
   }

}

/**
Generates report listing the items in the processors execution lists.
Each processor should have an identical list.
*/
void sys::report () {

processor* p = (processor*)lps[0];
unsigned num_exec = p->numExecuted ();

for (unsigned i = 0; i < lps.getSize (); i++) {
   std::cout << i << "\t\t\t";
   p = (processor*)lps[i];
   num_exec = std::min (num_exec, p->numExecuted());
   }

std::cout << std::endl;

message m, m1, m2;
for (unsigned i = 0; i < num_exec; i++) {
   for (unsigned j = 0; j < std::min (lps.getSize (), num_exec); j++) {
      processor* p = (processor*)lps[j];
      if (j == 0) 
         m = m1 = p->execList (i);
      else {
         m = m2 = p->execList (i);
         }
      std::cout << "(" << m.pid << "," << m.mid << "," << m.c << ")"; 
      std::cout << "\t\t\t";
      }
   std::cout << std::endl;
   }

}
 
sys::~sys () { 

lps.removeAll (); 

}

/**

Main function to start the simulator.

*/
int main (int argc, char** argv) {

sys* _sys = new sys (3, 10.0, 5.0, 10);

try {
   devssim::setThreadLimit (atoi (argv[1]));
   devssim sim (_sys);
   sim.reset ();
   sim.run ();
   _sys->report ();
   }
catch (DEVSException& err) { 
   std::cout << err.getError () << std::endl; 
   }

}
